/*
 * Framebuffer driver for SiS6326 compliant graphic boards
 * (C) 2001 Sergio Costas Rodriguez (raster@rastersoft.com)
 *
 * It's heavily based in VESA 2.0 FB driver
 * (which is (C) 1998 Gerd Knorr <kraxel@goldbach.in-berlin.de>)
 * That's why switching to graphics mode happens at boot time
 * (while running in real mode)
 */

// This is the version 1.0

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/console.h>
#include <linux/selection.h>
#include <linux/ioport.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/mtrr.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#include <video/fbcon-mac.h>

#define dac_reg	(0x3c8)
#define dac_val	(0x3c9)

/* --------------------------------------------------------------------- */

/*
 * card parameters
 */

/* card */
static unsigned long video_base; /* physical addr */
static int   video_size;
static char *video_vbase;        /* mapped */

/* mode */
static int  video_bpp;
static int  video_width;
static int  video_height;
static int  video_height_virtual;
static int  video_type = FB_TYPE_PACKED_PIXELS;
static int  video_visual;
static int  video_linelength;
static int  video_cmap_len;

struct sis6326fb_par {
    unsigned char misc;
    unsigned char feature;
    unsigned char enable;
    unsigned char segsel0;
    unsigned char segsel1;
    unsigned char crtindex[0x19];
    unsigned char seq[5];
    unsigned char graph[9];
    unsigned char attr[0x15];
    unsigned char xtended[0x3D];
};

/* --------------------------------------------------------------------- */

static struct fb_var_screeninfo sis6326fb_defined = {
    0,0,0,0,	/* W,H, W, H (virtual) load xres,xres_virtual*/
    0,0,		/* virtual -> visible no offset */
    8,		/* depth -> load bits_per_pixel */
    0,		/* greyscale ? */
    {0,0,0},	/* R */
    {0,0,0},	/* G */
    {0,0,0},	/* B */
    {0,0,0},	/* transparency */
    0,		/* standard pixel format */
    FB_ACTIVATE_NOW,
    -1,-1,
    0,
    0L,0L,0L,0L,0L,
    0L,0L,0,	/* No sync info */
    FB_VMODE_NONINTERLACED,
    {0,0,0,0,0,0}
};

static struct sis6326fb_par mode640x480x256={
    misc: 0xEF,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x5F,0x4F,0x4F,0x81,0x54,0x80,0xB,0x3E,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0xE1,0x4,0xDF,0x50,0x40,0xDF,0xFC,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x40,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x82,0xE0,0x3F,0x5,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x1B,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0xC1,0x14,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode640x480x32k={
    misc: 0xEF,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x5F,0x4F,0x4F,0x81,0x54,0x80,0xB,0x3E,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0xE1,0x84,0xDF,0xA0,0x4F,0xDF,0xFC,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x86,0x60,0x3F,0x5,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x1B,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode640x480x64k={
    misc: 0xEF,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x5F,0x4F,0x4F,0x81,0x54,0x80,0xB,0x3E,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0xE1,0x84,0xDF,0xA0,0x4F,0xDF,0xFC,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x8A,0x60,0x3F,0x5,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x1B,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode640x480x16m={
    misc: 0xEF,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x5F,0x4F,0x4F,0x81,0x54,0x80,0xB,0x3E,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0xE1,0x84,0xDF,0xF0,0x4F,0xDF,0xFC,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x92,0x60,0x4F,0x6,0x0,0x8C,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x1B,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};


static struct sis6326fb_par mode800x600x256={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x7B,0x63,0x63,0x1F,0x69,0x92,0x6F,0xF0,0x0,0x60,0x0,0x0,0x0,0x0,0x6D,0x20,0x58,0x8A,0x57,0x64,0x40,0x57,0x70,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x40,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x82,0x60,0x3F,0x5,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x43,0x48,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode800x600x32k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x7B,0x63,0x63,0x1F,0x69,0x92,0x6F,0xF0,0x0,0x60,0x0,0x0,0x0,0x0,0x6D,0x20,0x58,0x8A,0x57,0xC8,0x4F,0x57,0x70,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x86,0x60,0x4F,0x6,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x43,0x48,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode800x600x64k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x7B,0x63,0x63,0x1F,0x69,0x92,0x6F,0xF0,0x0,0x60,0x0,0x0,0x0,0x0,0x6D,0x20,0x58,0x8A,0x57,0xC8,0x4F,0x57,0x70,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x8A,0x60,0x4F,0x6,0x0,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x43,0x48,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode800x600x16m={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x7B,0x63,0x63,0x1F,0x69,0x92,0x6F,0xF0,0x0,0x60,0x0,0x0,0x0,0x0,0x6D,0x20,0x58,0x8A,0x57,0x2C,0x4F,0x57,0x70,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0x92,0x60,0x5F,0x7,0x10,0x8C,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x43,0x48,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};


static struct sis6326fb_par mode1024x768x256={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x99,0x7F,0x7F,0x1D,0x82,0x18,0x96,0x1F,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x7F,0x84,0x7F,0x0,0x40,0x7F,0x97,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x40,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xA2,0x60,0x3F,0x5,0x10,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x5D,0xC4,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1024x768x32k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x99,0x7F,0x7F,0x1D,0x82,0x18,0x96,0x1F,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x7F,0x84,0x7F,0x0,0x4F,0x7F,0x97,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xA6,0x60,0x5F,0x7,0x20,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x5D,0xC4,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1024x768x64k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x99,0x7F,0x7F,0x1D,0x82,0x18,0x96,0x1F,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x7F,0x84,0x7F,0x0,0x4F,0x7F,0x97,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xAA,0x60,0x5F,0x7,0x20,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x5D,0xC4,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1024x768x16m={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0x99,0x7F,0x7F,0x1D,0x82,0x18,0x96,0x1F,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x7F,0x84,0x7F,0x0,0x4F,0x7F,0x97,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xB2,0x60,0x6F,0x8,0x30,0x8C,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x40,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x5D,0xC4,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};


static struct sis6326fb_par mode1280x1024x256={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0xCF,0x9F,0x9F,0x93,0xB0,0x1F,0x14,0xBA,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x0,0x84,0xFF,0x40,0x40,0xFF,0x15,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x40,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xA2,0x60,0x4F,0x6,0x10,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x95,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1280x1024x32k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0xCF,0x9F,0x9F,0x93,0xB0,0x1F,0x14,0xBA,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x0,0x84,0xFF,0x80,0x4F,0xFF,0x15,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xA6,0x60,0x7F,0x9,0x20,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x95,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1280x1024x64k={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0xCF,0x9F,0x9F,0x93,0xB0,0x1F,0x14,0xBA,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x0,0x84,0xFF,0x80,0x4F,0xFF,0x15,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xAA,0x60,0x7F,0x9,0x20,0xC,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x95,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};

static struct sis6326fb_par mode1280x1024x16m={
    misc: 0x2F,
    feature: 0x1,
    enable: 0x1,
    segsel0: 0x0,
    segsel1: 0x0,
    crtindex: { 0xCF,0x9F,0x9F,0x93,0xB0,0x1F,0x14,0xBA,0x0,0x40,0x0,0x0,0x0,0x0,0x6D,0x20,0x0,0x84,0xFF,0xC0,0x4F,0xFF,0x15,0xA3,0xFF, },
    seq: { 0x3,0x1,0xF,0x0,0xE },
    graph: { 0x0,0x0,0x0,0x0,0x0,0x0,0x5,0xF,0xFF },
    attr: { 0x0,0x1,0x2,0x3,0x4,0x5,0x6,0x7,0x8,0x9,0xA,0xB,0xC,0xD,0xE,0xF,0x1,0x1,0x0,0x13,0x0 },
    xtended: { 0x0,0x0,0x0,0x0,0x0,0x0,0xB2,0x60,0x9F,0xB,0x30,0x8C,0xB6,0xF4,0x12,0x2,0x81,0x3,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x40,0x9B,0x0,0x2B,0x0,0x0,0x20,0x0,0x5A,0xE4,0x95,0xE1,0x0,0x10,0x0,0x0,0x0,0x0,0x0,0x80,0x81,0x4,0x0,0x80,0x0,0x0,0x0,0x0,0x40 }
};


static struct display disp;
static struct fb_info fb_info;
static struct { u_short blue, green, red, pad; } palette[256];
static int allow_set=0;
static union {
#ifdef FBCON_HAS_CFB16
    u16 cfb16[16];
#endif
#ifdef FBCON_HAS_CFB24
    u32 cfb24[16];
#endif
#ifdef FBCON_HAS_CFB32
    u32 cfb32[16];
#endif
} fbcon_cmap;

static int             inverse   = 0;
static int             mtrr      = 0;
static int             currcon   = 0;

static int             pmi_setpal = 0;	/* pmi for palette changes ??? */
static int             ypan       = 0;  /* 0..nothing, 1..ypan, 2..ywrap */
static unsigned short  *pmi_base  = 0;
static void            (*pmi_start)(void);
static void            (*pmi_pal)(void);

static struct display_switch sis6326fb_sw;

/* --------------------------------------------------------------------- */


static void sis6326fb_change_mode(struct sis6326fb_par *par) {
    
    /*
     *  Set the hardware according to 'par'.
     */
    
    int direccion,bucle;
    unsigned char temporal,datoand;
    
    if(((par->misc) & 1)==0)
	direccion=0xB0;
    else
	direccion=0xD0;
    
    outb(((par->misc) & 0xEF),0x3c2);
    outb(((par->feature) & 0x08),0x30A+direccion);
    outb(((par->enable) & 0x01),0x3C3);
    outb(par->segsel0,0x3CD);
    outb(par->segsel1,0x3CB);
    
    temporal=par->crtindex[0x11];
    temporal=temporal&0x7F;
    outb(0x11,0x304+direccion);
    outb(temporal,0x305+direccion); // disable write protection
    
    for(bucle=0;bucle<0x19;bucle++) {
	
	switch(bucle) {
	case 3:
	    datoand=0x7F;
	    break;
	case 8:
	    datoand=0x7F;
	    break;
	case 0xA:
	    datoand=0x3F;
	    break;
	case 0xB:
	    datoand=0x7F;
	    break;
	case 0x14:
	    datoand=0x7F;
	    break;
	case 0x17:
	    datoand=0xEF;
	    break;
	default:
	    datoand=0xFF;
	}
	outb(bucle,0x304+direccion);
	outb(((par->crtindex[bucle]) & datoand),0x305+direccion);
    }
    
    for(bucle=0;bucle<0x5;bucle++) {
	
	switch (bucle) {
	case 0x0:
	    datoand=0x03;
	    break;
	case 0x1:
	    datoand=0x3D;
	    break;
	case 0x2:
	    datoand=0x0F;
	    break;
	case 0x3:
	    datoand=0x3F;
	    break;
	case 0x4:
	    datoand=0x0E;
	    break;
	default:
	    datoand=0xFF;
	}
	
	outb(bucle,0x3C4);
	outb(((par->seq[bucle]) & datoand),0x3C5);
    }
    
    for(bucle=0;bucle<0x9;bucle++) {
	
	switch (bucle) {
	case 0x0:
	    datoand=0x0F;
	    break;
	case 0x1:
	    datoand=0x0F;
	    break;
	case 0x2:
	    datoand=0x0F;
	    break;
	case 0x3:
	    datoand=0x1F;
	    break;
	case 0x4:
	    datoand=0x03;
	    break;
	case 0x5:
	    datoand=0x7D;
	    break;
	case 0x6:
	    datoand=0x0F;
	    break;
	case 0x7:
	    datoand=0x0F;
	    break;
	default:
	    datoand=0xFF;
	}
	
	outb(bucle,0x3CE);
	outb(((par->graph[bucle]) & datoand),0x3CF);
    }
    
    temporal=inb(0x30A+direccion);
    
    for(bucle=0;bucle<0x15;bucle++) {
	
	switch (bucle) {
	case 0x0:
	    datoand=0x3F;
	    break;
	case 0x1:
	    datoand=0x3F;
	    break;
	case 0x2:
	    datoand=0x3F;
	    break;
	case 0x3:
	    datoand=0x3F;
	    break;
	case 0x4:
	    datoand=0x3F;
	    break;
	case 0x5:
	    datoand=0x3F;
	    break;
	case 0x6:
	    datoand=0x3F;
	    break;
	case 0x7:
	    datoand=0x3F;
	    break;
	case 0x8:
	    datoand=0x3F;
	    break;
	case 0x9:
	    datoand=0x3F;
	    break;
	case 0xA:
	    datoand=0x3F;
	    break;
	case 0xB:
	    datoand=0x3F;
	    break;
	case 0xC:
	    datoand=0x3F;
	    break;
	case 0xD:
	    datoand=0x3F;
	    break;
	case 0xE:
	    datoand=0x3F;
	    break;
	case 0xF:
	    datoand=0x3F;
	    break;
	case 0x10:
	    datoand=0xEF;
	    break;
	case 0x11:
	    datoand=0x3F;
	    break;
	case 0x12:
	    datoand=0x3F;
	    break;
	case 0x13:
	    datoand=0x0F;
	    break;
	case 0x14:
	    datoand=0x0F;
	    break;
	default:
	    datoand=0xFF;
	}
	
	outb(bucle,0x3C0);
	outb(((par->attr[bucle]) & datoand),0x3C1);
    }
    
    outb(5,0x3C4); //unlock extended registers
    outb(0x86,0x3c5);
    temporal=inb(0x3c5);
    
    for(bucle=6;bucle<0x3D;bucle++) {
	
	switch (bucle) {
	case 0xC:
	    datoand=0xEF;
	    break;
	case 0x11:
	    datoand=0xF0;
	    break;
	case 0x13:
	    datoand=0xC0;
	    break;
	case 0x14:
	    datoand=0x3F;
	    break;
	case 0x15:
	    datoand=0x3F;
	    break;
	case 0x16:
	    datoand=0x3F;
	    break;
	case 0x17:
	    datoand=0x3F;
	    break;
	case 0x18:
	    datoand=0x3F;
	    break;
	case 0x19:
	    datoand=0x3F;
	    break;
	case 0x1B:
	    datoand=0x07;
	    break;
	case 0x1C:
	    datoand=0x3F;
	    break;
	case 0x1F:
	    datoand=0x3F;
	    break;
	case 0x21:
	    datoand=0x7F;
	    break;
	case 0x23:
	    datoand=0x7F;
	    break;
	case 0x26:
	    datoand=0x77;
	    break;
	case 0x2C:
	    datoand=0x7F;
	    break;
	case 0x2D:
	    datoand=0x0F;
	    break;
	case 0x2F:
	    datoand=0x30;
	    break;
	case 0x32:
	    datoand=0x0F;
	    break;
	case 0x33:
	    datoand=0x7F;
	    break;
	case 0x34:
	    datoand=0xC5;
	    break;
	case 0x38:
	    datoand=0xF7;
	    break;
	case 0x39:
	    datoand=0x1F;
	    break;
	case 0x3A:
	    datoand=0x7F;
	    break;
	case 0x3C:
	    datoand=0x7F;
	    break;
	default:
	    datoand=0xFF;
	}
	
	if((bucle!=0x0D) && (bucle!=0x0E) && (bucle!=0x0F) && (bucle!=0x10) &&
	   (bucle!=0x20) && (bucle!=0x21) && (bucle!=0x23) && (bucle!=0x24) &&
	   (bucle!=0x25) && (bucle!=0x26) && (bucle!=0x27) && (bucle!=0x2C) &&
	   (bucle!=0x2D) && (bucle!=0x2E) && (bucle!=0x2F) && (bucle!=0x30) &&
	   (bucle!=0x31) && (bucle!=0x32) && (bucle!=0x33) && (bucle!=0x34) &&
	   (bucle!=0x35) && (bucle!=0x36) && (bucle!=0x37) && (bucle!=0x39) && 
	   (bucle!=0x3A) && (bucle!=0x3C)) {
	    
	    outb(bucle,0x3C4);
	    outb(((par->xtended[bucle]) & datoand),0x3C5);
	}
    }
    
    outb(5,0x3C4);
    outb(0,0x3C5);
    temporal=inb(0x3C5); // lock extended registers
}


/* --------------------------------------------------------------------- */

static int sis6326fb_pan_display(struct fb_var_screeninfo *var, int con,
                              struct fb_info *info)
{
    int offset;
    
    if (!ypan)
	return -EINVAL;
    if (var->xoffset)
	return -EINVAL;
    if (var->yoffset > var->yres_virtual)
	return -EINVAL;
    if ((ypan==1) && var->yoffset+var->yres > var->yres_virtual)
	return -EINVAL;
    
    offset = (var->yoffset * video_linelength + var->xoffset) / 4;
    
    __asm__ __volatile__(
			 "call *(%%edi)"
			 : /* no return value */
			 : "a" (0x4f07),         /* EAX */
			 "b" (0),              /* EBX */
			 "c" (offset),         /* ECX */
			 "d" (offset >> 16),   /* EDX */
			 "D" (&pmi_start));    /* EDI */
    return 0;
}

static int sis6326fb_update_var(int con, struct fb_info *info)
{
    if (con == currcon && ypan) {
	struct fb_var_screeninfo *var = &fb_display[currcon].var;
	return sis6326fb_pan_display(var,con,info);
    }
    return 0;
}

static int sis6326fb_get_fix(struct fb_fix_screeninfo *fix, int con,
			     struct fb_info *info)
{
    memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    strcpy(fix->id,"SiS6326 VGA");
    
    fix->smem_start=video_base;
    fix->smem_len=video_size;
    fix->type = video_type;
    fix->visual = video_visual;
    fix->xpanstep  = 0;
    fix->ypanstep  = ypan     ? 1 : 0;
    fix->ywrapstep = (ypan>1) ? 1 : 0;
    fix->line_length=video_linelength;
    return 0;
}

static int sis6326fb_get_var(struct fb_var_screeninfo *var, int con,
			     struct fb_info *info) {
    if(con==-1) {
	memcpy(var, &sis6326fb_defined, sizeof(struct fb_var_screeninfo));
	var->xres=video_width;
	var->yres=video_height;
	var->xres_virtual=video_width;
	var->yres_virtual=video_height_virtual;
	var->xoffset=0;
	var->yoffset=0;
	var->grayscale=0;
	var->bits_per_pixel=video_bpp;
	var->nonstd=0;
	var->accel_flags=0;

	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;

	switch(var->bits_per_pixel) {
	case 8:
	    var->red.offset=0;
	    var->red.length = 8;
	    var->green.offset = 0;
	    var->green.length = 8;
	    var->blue.offset = 0;
	    var->blue.length = 8;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	case 15:
	    var->red.offset=10;
	    var->red.length = 5;
	    var->green.offset = 5;
	    var->green.length = 5;
	    var->blue.offset = 0;
	    var->blue.length = 5;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	case 16:
	    var->red.offset=11;
	    var->red.length = 5;
	    var->green.offset = 5;
	    var->green.length = 6;
	    var->blue.offset = 0;
	    var->blue.length = 5;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	case 24:
	    var->red.offset=16;
	    var->red.length = 8;
	    var->green.offset = 8;
	    var->green.length = 8;
	    var->blue.offset = 0;
	    var->blue.length = 8;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	}
    } else
	*var=fb_display[con].var;
    return 0;
}

static void sis6326fb_set_disp(int con)
{
    struct fb_fix_screeninfo fix;
    struct display *display;
    struct display_switch *sw;
    
    if (con >= 0)
	display = &fb_display[con];
    else
	display = &disp;	/* used during initialization */
    
    sis6326fb_get_fix(&fix, con, 0);
    
    memset(display, 0, sizeof(struct display));
    display->screen_base = video_vbase;
    display->visual = fix.visual;
    display->type = fix.type;
    display->type_aux = fix.type_aux;
    display->ypanstep = fix.ypanstep;
    display->ywrapstep = fix.ywrapstep;
    display->line_length = fix.line_length;
    display->next_line = fix.line_length;
    display->can_soft_blank = 0;
    display->inverse = inverse;
    sis6326fb_get_var(&display->var, -1, &fb_info);
    
    switch (video_bpp) {
    case 8:
	sw = &fbcon_cfb8;
	break;
    case 15:
    case 16:
	sw = &fbcon_cfb16;
	display->dispsw_data = fbcon_cmap.cfb16;
	break;
    case 24:
	sw = &fbcon_cfb24;
	display->dispsw_data = fbcon_cmap.cfb24;
	break;
    default:
#ifdef FBCON_HAS_MAC
	sw = &fbcon_mac;
	break;
#else
	sw = &fbcon_dummy;
	return;
#endif
    }
    memcpy(&sis6326fb_sw, sw, sizeof(struct display_switch));
    if (!ypan) {
	display->scrollmode = SCROLL_YREDRAW;
	sis6326fb_sw.bmove = fbcon_redraw_bmove;
    }
    display->dispsw = &sis6326fb_sw;
}

static int sis6326fb_set_var2(struct fb_var_screeninfo *var, int con,
			      struct fb_info *info) {
    
    struct display *display;
    struct sis6326fb_par *parametros;
    int yresvirtualmax=0;
    
    if(con>=0)
	display=&fb_display[con];
    else
	display=&disp;
    
    switch(var->bits_per_pixel) {
    case 8:
	yresvirtualmax=video_size/var->xres;
	break;
    case 15:
    case 16:
	yresvirtualmax=video_size/(2*var->xres);
	break;
    case 24:
	yresvirtualmax=video_size/(3*var->xres);
	break;
    default:
	yresvirtualmax=var->yres;
    }
    
    
    if(((var->xres!=640)&&(var->xres!=800)&&(var->xres!=1024)&&(var->xres!=1280))||
       ((var->yres!=480)&&(var->yres!=600)&&(var->yres!=768)&&(var->yres!=1024)) ||
       (var->xres_virtual!=var->xres) ||
       (var->yres_virtual != var->yres) ||
       var->xoffset ||
       ((var->bits_per_pixel !=8)&&(var->bits_per_pixel !=15)&&(var->bits_per_pixel !=16)&&(var->bits_per_pixel !=24)) ||
       var->nonstd ||
       ((var->xres==640)&&(var->yres!=480)) ||
       ((var->xres==800)&&(var->yres!=600)) ||
       ((var->xres==1024)&&(var->yres!=768)) ||
       ((var->xres==1280)&&(var->yres!=1024)))
	
	return -EINVAL;
    
    if ((var->yoffset)&& !ypan)
	return -EINVAL;
    
    if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_TEST)
	return 0;
    
    if(con != currcon)
	return 0;
    
    parametros=NULL;
    
    if((var->xres==640)&&(var->yres==480)&&(var->bits_per_pixel==8)) {
	parametros=&mode640x480x256;
	video_bpp=8;
	video_width=640;
	video_height=480;
	video_visual=FB_VISUAL_PSEUDOCOLOR;
	video_linelength=640;
	video_cmap_len=256;
    }
    
    if((var->xres==800)&&(var->yres==600)&&(var->bits_per_pixel==8)) {
	parametros=&mode800x600x256;
	video_bpp=8;
	video_width=800;
	video_height=600;
	video_visual=FB_VISUAL_PSEUDOCOLOR;
	video_linelength=800;
	video_cmap_len=256;
    }
    
    if((var->xres==1024)&&(var->yres==768)&&(var->bits_per_pixel==8)) {
	parametros=&mode1024x768x256;
	video_bpp=8;
	video_width=1024;
	video_height=768;
	video_visual=FB_VISUAL_PSEUDOCOLOR;
	video_linelength=1024;
	video_cmap_len=256;
    }
    
    if((var->xres==1280)&&(var->yres==1024)&&(var->bits_per_pixel==8)) {
	parametros=&mode1280x1024x256;
	video_bpp=8;
	video_width=1280;
	video_height=1024;
	video_visual=FB_VISUAL_PSEUDOCOLOR;
	video_linelength=1280;
	video_cmap_len=256;
    }
    
    
    if((var->xres==640)&&(var->yres==480)&&(var->bits_per_pixel==15)) {
	parametros=&mode640x480x32k;
	video_bpp=15;
	video_width=640;
	video_height=480;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=1280;
	video_cmap_len=16;
    }
    
    if((var->xres==800)&&(var->yres==600)&&(var->bits_per_pixel==15)) {
	parametros=&mode800x600x32k;
	video_bpp=15;
	video_width=800;
	video_height=600;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=1600;
	video_cmap_len=16;
    }
    
    if((var->xres==1024)&&(var->yres==768)&&(var->bits_per_pixel==15)) {
	parametros=&mode1024x768x32k;
	video_bpp=15;
	video_width=1024;
	video_height=768;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=2048;
	video_cmap_len=16;
    }
    
    if((var->xres==1280)&&(var->yres==1024)&&(var->bits_per_pixel==15)) {
	parametros=&mode1280x1024x32k;
	video_bpp=15;
	video_width=1280;
	video_height=1024;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=2560;
	video_cmap_len=16;
    }
    
    
    if((var->xres==640)&&(var->yres==480)&&(var->bits_per_pixel==16)) {
	parametros=&mode640x480x64k;
	video_bpp=16;
	video_width=640;
	video_height=480;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=1280;
	video_cmap_len=16;
    }
    
    if((var->xres==800)&&(var->yres==600)&&(var->bits_per_pixel==16)) {
	parametros=&mode800x600x64k;
	video_bpp=16;
	video_width=800;
	video_height=600;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=1600;
	video_cmap_len=16;
    }
    
    if((var->xres==1024)&&(var->yres==768)&&(var->bits_per_pixel==16)) {
	parametros=&mode1024x768x64k;
	video_bpp=16;
	video_width=1024;
	video_height=768;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=2048;
	video_cmap_len=16;
    }
    
    if((var->xres==1280)&&(var->yres==1024)&&(var->bits_per_pixel==16)) {
	parametros=&mode1280x1024x64k;
	video_bpp=16;
	video_width=1280;
	video_height=1024;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=2560;
	video_cmap_len=16;
    }
    
    
    if((var->xres==640)&&(var->yres==480)&&(var->bits_per_pixel==24)) {
	parametros=&mode640x480x16m;
	video_bpp=24;
	video_width=640;
	video_height=480;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=1920;
	video_cmap_len=16;
    }
    
    if((var->xres==800)&&(var->yres==600)&&(var->bits_per_pixel==24)) {
	parametros=&mode800x600x16m;
	video_bpp=24;
	video_width=800;
	video_height=600;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=2400;
	video_cmap_len=16;
    }
    
    if((var->xres==1024)&&(var->yres==768)&&(var->bits_per_pixel==24)) {
	parametros=&mode1024x768x16m;
	video_bpp=24;
	video_width=1024;
	video_height=768;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=3072;
	video_cmap_len=16;
    }
    
    if((var->xres==1280)&&(var->yres==1024)&&(var->bits_per_pixel==24)) {
	parametros=&mode1280x1024x16m;
	video_bpp=24;
	video_width=1280;
	video_height=1024;
	video_visual=FB_VISUAL_TRUECOLOR;
	video_linelength=3840;
	video_cmap_len=16;
    }
    
    
    if(parametros==NULL)
	return -EINVAL;
    
    video_height_virtual=video_size/video_linelength;
    
    sis6326fb_change_mode(parametros);
    return 0;
}

static int sis6326fb_set_var(struct fb_var_screeninfo *var, int con,
			  struct fb_info *info)
{
    
    int retorno;
    struct display *display;
    struct fb_fix_screeninfo fix;	
    
    allow_set=1;

    retorno=sis6326fb_set_var2(var, con, info);
    if(retorno)
	return(retorno);

    if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_TEST) {
	var->vmode=FB_VMODE_CONUPDATE;
	return 0;
    }
    
    if(con<0)
	return 0;
    display=&fb_display[con];

    display->var=*var;
    display->var.vmode=FB_VMODE_CONUPDATE;
    display->screen_base=video_vbase;
    sis6326fb_get_fix(&fix,-1,NULL);
    display->visual=fix.visual;
    display->type = fix.type;
    display->type_aux = fix.type_aux;
    display->ypanstep = fix.ypanstep;
    display->ywrapstep = fix.ywrapstep;
    display->line_length = fix.line_length;
    display->next_line=fix.line_length;
    display->can_soft_blank = 1;
    display->inverse = 0;
    
    if(currcon!=con) {
	switch(display->var.bits_per_pixel) {
	case 8:
	    display->visual=FB_VISUAL_PSEUDOCOLOR;
	    display->line_length=var->xres_virtual;
	    display->next_line=var->xres_virtual;
	    break;
	case 15:
	    display->visual=FB_VISUAL_TRUECOLOR;
	    display->line_length=2*(var->xres_virtual);
	    display->next_line=2*(var->xres_virtual);
	    break;	
	case 16:
	    display->visual=FB_VISUAL_TRUECOLOR;
	    display->line_length=2*(var->xres_virtual);
	    display->next_line=2*(var->xres_virtual);
	    break;	
	case 24:
	    display->visual=FB_VISUAL_TRUECOLOR;
	    display->line_length=3*(var->xres_virtual);
	    display->next_line=3*(var->xres_virtual);
	    break;
	default:
	}
    }

    display->var.red.msb_right = 0;
    display->var.green.msb_right = 0;
    display->var.blue.msb_right = 0;
    display->var.transp.msb_right = 0;

    switch(display->var.bits_per_pixel) {
    case 8:
	display->dispsw = &fbcon_cfb8;
	display->var.red.offset=0;
	display->var.red.length = 8;
	display->var.green.offset = 0;
	display->var.green.length = 8;
	display->var.blue.offset = 0;
	display->var.blue.length = 8;
	display->var.transp.offset = 0;
	display->var.transp.length = 0;
	break;
    case 15:
	display->dispsw = &fbcon_cfb16;
	display->dispsw_data = fbcon_cmap.cfb16;
	display->var.red.offset=10;
	display->var.red.length = 5;
	display->var.green.offset = 5;
	display->var.green.length = 5;
	display->var.blue.offset = 0;
	display->var.blue.length = 5;
	display->var.transp.offset = 0;
	display->var.transp.length = 0;
	break;
    case 16:
	display->dispsw = &fbcon_cfb16;
	display->dispsw_data = fbcon_cmap.cfb16;
	display->var.red.offset=11;
	display->var.red.length = 5;
	display->var.green.offset = 5;
	display->var.green.length = 6;
	display->var.blue.offset = 0;
	display->var.blue.length = 5;
	display->var.transp.offset = 0;
	display->var.transp.length = 0;
	break;
    case 24:
	display->dispsw = &fbcon_cfb24;
	display->dispsw_data = fbcon_cmap.cfb24;
	display->var.red.offset=16;
	display->var.red.length = 8;
	display->var.green.offset = 8;
	display->var.green.length = 8;
	display->var.blue.offset = 0;
	display->var.blue.length = 8;
	display->var.transp.offset = 0;
	display->var.transp.length = 0;
	break;
    }
    
    if (fb_info.changevar)
	(*fb_info.changevar)(con);

    return 0;
}

static int sis6326_getcolreg(unsigned regno, unsigned *red, unsigned *green,
			     unsigned *blue, unsigned *transp,
			     struct fb_info *fb_info)
{
    /*
     *  Read a single color register and split it into colors/transparent.
     *  Return != 0 for invalid regno.
     */
    
    if (regno >= video_cmap_len)
	return 1;
    
    *red   = palette[regno].red;
    *green = palette[regno].green;
    *blue  = palette[regno].blue;
    *transp = 0;
    return 0;
}

static void sis6326_setpalette(int regno, unsigned red, unsigned green, unsigned blue)
{
    struct { u_char blue, green, red, pad; } entry;
    
    if (pmi_setpal) {
	entry.red   = red   >> 10;
	entry.green = green >> 10;
	entry.blue  = blue  >> 10;
	entry.pad   = 0;
	__asm__ __volatile__(
			     "call *(%%esi)"
			     : /* no return value */
			     : "a" (0x4f09),         /* EAX */
			     "b" (0),              /* EBX */
			     "c" (1),              /* ECX */
			     "d" (regno),          /* EDX */
			     "D" (&entry),         /* EDI */
			     "S" (&pmi_pal));      /* ESI */
    } else {
	/* without protected mode interface, try VGA registers... */
	outb_p(regno,       dac_reg);
	outb_p(red   >> 10, dac_val);
	outb_p(green >> 10, dac_val);
	outb_p(blue  >> 10, dac_val);
    }
}

static int sis6326_setcolreg(unsigned regno, unsigned red, unsigned green,
			     unsigned blue, unsigned transp,
			     struct fb_info *fb_info)
{
    /*
     *  Set a single color register. The values supplied are
     *  already rounded down to the hardware's capabilities
     *  (according to the entries in the `var' structure). Return
     *  != 0 for invalid regno.
     */
    
    if (regno >= video_cmap_len)
	return 1;
    
    palette[regno].red   = red;
    palette[regno].green = green;
    palette[regno].blue  = blue;
    
    switch (video_bpp) {
    case 8:
	sis6326_setpalette(regno,red,green,blue);
	break;
    case 15:
	fbcon_cmap.cfb16[regno] =
	    ((red   & 0xf800) >>  1) |
	    ((green & 0xf800) >>  6) |
	    ((blue  & 0xf800) >> 11);
	break;
    case 16:
	fbcon_cmap.cfb16[regno] =
	    ((red   & 0xf800)      ) |
	    ((green & 0xfc00) >>  5) |
	    ((blue  & 0xf800) >> 11);
	break;
    case 24:
	red   >>= 8;
	green >>= 8;
	blue  >>= 8;
	fbcon_cmap.cfb24[regno] =
	    (red   << 16)   |
	    (green << 8) |
	    (blue);
	break;
    }
    return 0;
}

static void do_install_cmap(int con, struct fb_info *info)
{
    if (con != currcon)
	return;
    if (fb_display[con].cmap.len)
	fb_set_cmap(&fb_display[con].cmap, 1, sis6326_setcolreg, info);
    else
	fb_set_cmap(fb_default_cmap(video_cmap_len), 1, sis6326_setcolreg,
		    info);
}

static int sis6326fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			      struct fb_info *info)
{
    if (con == currcon) /* current console? */
	return fb_get_cmap(cmap, kspc, sis6326_getcolreg, info);
    else if (fb_display[con].cmap.len) /* non default colormap? */
	fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
    else
	fb_copy_cmap(fb_default_cmap(video_cmap_len),
		     cmap, kspc ? 0 : 2);
    return 0;
}

static int sis6326fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			      struct fb_info *info)
{
    int err;
    
    if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
	err = fb_alloc_cmap(&fb_display[con].cmap,video_cmap_len,0);
	if (err)
	    return err;
    }
    if (con == currcon)			/* current console? */
	return fb_set_cmap(cmap, kspc, sis6326_setcolreg, info);
    else
	fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
    return 0;
}

static struct fb_ops sis6326fb_ops = {
    owner:		THIS_MODULE,
    fb_get_fix:	sis6326fb_get_fix,
    fb_get_var:	sis6326fb_get_var,
    fb_set_var:	sis6326fb_set_var,
    fb_get_cmap:	sis6326fb_get_cmap,
    fb_set_cmap:	sis6326fb_set_cmap,
    fb_pan_display:	sis6326fb_pan_display,
};

int __init sis6326fb_setup(char *options)
{
    char *this_opt;
    
    fb_info.fontname[0] = '\0';

    if (!options || !*options)
	return 0;

    for(this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) {
	if (!*this_opt) continue;
	
	if (! strcmp(this_opt, "inverse"))
	    inverse=1;
	else if (! strcmp(this_opt, "redraw"))
	    ypan=0;
	else if (! strcmp(this_opt, "ypan"))
	    ypan=1;
	else if (! strcmp(this_opt, "ywrap"))
	    ypan=2;
	else if (! strcmp(this_opt, "vgapal"))
	    pmi_setpal=0;
	else if (! strcmp(this_opt, "pmipal"))
	    pmi_setpal=1;
	else if (! strcmp(this_opt, "mtrr"))
	    mtrr=1;
	else if (!strncmp(this_opt, "font:", 5))
	    strcpy(fb_info.fontname, this_opt+5);
    }
    return 0;
}

static int sis6326fb_switch(int con, struct fb_info *info) {
    
    static int inicio=0;
    
    if(con==currcon)
	return 0;

    /* Do we have to save the colormap? */
    if (fb_display[currcon].cmap.len)
	fb_get_cmap(&fb_display[currcon].cmap, 1, sis6326_getcolreg,
		    info);
    
    currcon = con;

    if(allow_set) // only allow to change the graphic mode if we already has made it.
	sis6326fb_set_var2(&fb_display[con].var,con,info); // this avoids a hang while booting
    
    inicio=1;
    do_install_cmap(con, info);

    return 0;
}

/* 0 unblank, 1 blank, 2 no vsync, 3 no hsync, 4 off */

static void sis6326fb_blank(int blank, struct fb_info *info)
{
    /* Not supported... yet */
}

int __init sis6326fb_init(void)
{
    int i,j;
    
    if (screen_info.orig_video_isVGA != VIDEO_TYPE_VLFB)
	return -ENXIO;
    
    video_base          = screen_info.lfb_base;
    video_bpp           = screen_info.lfb_depth;
    if (15 == video_bpp)
	video_bpp = 16;
    video_width         = screen_info.lfb_width;
    video_height        = screen_info.lfb_height;
    video_linelength    = screen_info.lfb_linelength;
    video_size          = screen_info.lfb_size * 65536;
    video_visual = (video_bpp == 8) ?
	FB_VISUAL_PSEUDOCOLOR : FB_VISUAL_TRUECOLOR;
    
    if (!request_mem_region(video_base, video_size, "sis6326fb")) {
	printk(KERN_ERR
	       "sis6326fb: abort, cannot reserve video memory at 0x%lx\n",
	       video_base);
	return -EBUSY;
    }
    
    video_vbase = ioremap(video_base, video_size);
    if (!video_vbase) {
	release_mem_region(video_base, video_size);
	printk(KERN_ERR
	       "sis6326fb: abort, cannot ioremap video memory 0x%x @ 0x%lx\n",
	       video_size, video_base);
	return -EIO;
    }
    
    printk(KERN_INFO "sis6326fb: framebuffer at 0x%lx, mapped to 0x%p, size %dk\n",
	   video_base, video_vbase, video_size/1024);
    printk(KERN_INFO "sis6326fb: mode is %dx%dx%d, linelength=%d, pages=%d\n",
	   video_width, video_height, video_bpp, video_linelength, screen_info.pages);
    
    if (screen_info.vesapm_seg) {
	printk(KERN_INFO "sis6326fb: protected mode interface info at %04x:%04x\n",
	       screen_info.vesapm_seg,screen_info.vesapm_off);
    }
    
    if (screen_info.vesapm_seg < 0xc000)
	ypan = pmi_setpal = 0; /* not available or some DOS TSR ... */
    
    if (ypan || pmi_setpal) {
	pmi_base  = (unsigned short*)bus_to_virt(((unsigned long)screen_info.vesapm_seg << 4) + screen_info.vesapm_off);
	pmi_start = (void*)((char*)pmi_base + pmi_base[1]);
	pmi_pal   = (void*)((char*)pmi_base + pmi_base[2]);
	printk(KERN_INFO "sis6326fb: pmi: set display start = %p, set palette = %p\n",pmi_start,pmi_pal);
	if (pmi_base[3]) {
	    printk(KERN_INFO "sis6326fb: pmi: ports = ");
	    for (i = pmi_base[3]/2; pmi_base[i] != 0xffff; i++)
		printk("%x ",pmi_base[i]);
	    printk("\n");
	    if (pmi_base[i] != 0xffff) {
				/*
				 * memory areas not supported (yet?)
				 *
				 * Rules are: we have to set up a descriptor for the requested
				 * memory area and pass it in the ES register to the BIOS function.
				 */
		printk(KERN_INFO "sis6326fb: can't handle memory requests, pmi disabled\n");
		ypan = pmi_setpal = 0;
	    }
	}
    }
    
    sis6326fb_defined.xres=video_width;
    sis6326fb_defined.yres=video_height;
    sis6326fb_defined.xres_virtual=video_width;
    sis6326fb_defined.yres_virtual=video_size / video_linelength;
    sis6326fb_defined.bits_per_pixel=video_bpp;
    
    if (ypan && sis6326fb_defined.yres_virtual > video_height) {
	printk(KERN_INFO "sis6326fb: scrolling: %s using protected mode interface, yres_virtual=%d\n",
	       (ypan > 1) ? "ywrap" : "ypan",sis6326fb_defined.yres_virtual);
    } else {
	printk(KERN_INFO "sis6326fb: scrolling: redraw\n");
	sis6326fb_defined.yres_virtual = video_height;
	ypan = 0;
    }
    video_height_virtual = sis6326fb_defined.yres_virtual;
    
    /* some dummy values for timing to make fbset happy */
    sis6326fb_defined.pixclock     = 10000000 / video_width * 1000 / video_height;
    sis6326fb_defined.left_margin  = (video_width / 8) & 0xf8;
    sis6326fb_defined.right_margin = 32;
    sis6326fb_defined.upper_margin = 16;
    sis6326fb_defined.lower_margin = 4;
    sis6326fb_defined.hsync_len    = (video_width / 8) & 0xf8;
    sis6326fb_defined.vsync_len    = 4;
    
    if (video_bpp > 8) {
	sis6326fb_defined.red.offset    = screen_info.red_pos;
	sis6326fb_defined.red.length    = screen_info.red_size;
	sis6326fb_defined.green.offset  = screen_info.green_pos;
	sis6326fb_defined.green.length  = screen_info.green_size;
	sis6326fb_defined.blue.offset   = screen_info.blue_pos;
	sis6326fb_defined.blue.length   = screen_info.blue_size;
	sis6326fb_defined.transp.offset = screen_info.rsvd_pos;
	sis6326fb_defined.transp.length = screen_info.rsvd_size;
	printk(KERN_INFO "sis6326fb: directcolor: "
	       "size=%d:%d:%d:%d, shift=%d:%d:%d:%d\n",
	       screen_info.rsvd_size,
	       screen_info.red_size,
	       screen_info.green_size,
	       screen_info.blue_size,
	       screen_info.rsvd_pos,
	       screen_info.red_pos,
	       screen_info.green_pos,
	       screen_info.blue_pos);
	video_cmap_len = 16;
    } else {
	sis6326fb_defined.red.length   = 6;
	sis6326fb_defined.green.length = 6;
	sis6326fb_defined.blue.length  = 6;
	for(i = 0; i < 16; i++) {
	    j = color_table[i];
	    palette[i].red   = default_red[j];
	    palette[i].green = default_grn[j];
	    palette[i].blue  = default_blu[j];
	}
	video_cmap_len = 256;
    }
    
    /* request failure does not faze us, as vgacon probably has this
     * region already (FIXME) */
    request_region(0x3c0, 32, "sis6326fb");
    
    if (mtrr) {
	int temp_size = video_size;
	while (mtrr_add(video_base, temp_size, MTRR_TYPE_WRCOMB, 1)==-EINVAL) {
	    temp_size >>= 1;
	}
    }
    
    strcpy(fb_info.modename, "SiS6326 VGA");
    fb_info.changevar = NULL;
    fb_info.node = -1;
    fb_info.fbops = &sis6326fb_ops;
    fb_info.disp=&disp;
    fb_info.switch_con=&sis6326fb_switch;
    fb_info.updatevar=&sis6326fb_update_var;
    fb_info.blank=&sis6326fb_blank;
    fb_info.flags=FBINFO_FLAG_DEFAULT;
    sis6326fb_set_disp(-1);
    
    if (register_framebuffer(&fb_info)<0)
	return -EINVAL;
    
    printk(KERN_INFO "fb%d: %s frame buffer device\n",
	       GET_FB_IDX(fb_info.node), fb_info.modename);
    return 0;
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
